from .symbol import ZeroLiteral, Literal, Symbol, ListRef, List, IgnoredGradient
from .stdlib import add

class SymbolCollection(dict):
    """ A dictionary to look up collected symbols generated
        during the autodiff
    """
    def __init__(self, model, tape):
        self.model = model
        self.zero = ZeroLiteral()
        self._max_ref_id = {}

        def _visit_ref(ref):
            if isinstance(ref, ListRef):
                for r1 in ref:
                    _visit_ref(r1)
            elif isinstance(ref.symbol, Symbol):
                self._max_ref_id[ref.symbol._name] = max(
                    self._max_ref_id.get(ref.symbol._name, 0), ref.ref_id
                    )

        for record in tape:
            node = record.node
            for ref in node.varin.values():
                _visit_ref(ref)


    def add(self, var):
        self[var._name] = var
        return var

    def add_vjp(self, var, ref_id=None):
        if ref_id is None:
            var_p = Symbol(var._vjp_name, model=self.model)
        # there can be many versions of vjp with distinct values.
        else:
            # first occurance, must have the max ref_id.
            var_p = Symbol(var._vjp_name + '#%d' % ref_id, model=self.model)
        return self.add(var_p)

    def add_jvp(self, var):
        return self.add(Symbol(var._jvp_name, model=self.model))

    def is_last_ref(self, var, ref_id):
        if var._name not in self._max_ref_id or ref_id == self._max_ref_id[var._name]:
            return True

        return False

    def get_vjp(self, var, ref_id=None):
        if ref_id is None:
            return self[var._vjp_name]
        else:
            return self[var._vjp_name + '#%d' % ref_id]

    def get_jvp(self, var):
        return self[var._jvp_name]

    def check(self, model):
        for var in self.values():
            assert var._model is model

def prepare_opr_kwargs(record, model):
    """ generate a first guess of kwargs based on the record.

    """
    p = record.node
    impl_kwargs = record.impl_kwargs

    kwargs = {}
    kwargs.update(impl_kwargs)
    return kwargs

def create_output_vjp(ref, symbols):

    # make lists for lists
    if isinstance(ref, ListRef):
        return List([
                    create_output_vjp(r, symbols)
                    for r in ref]
                )

    var = ref.symbol

    # bypass literal arguments
    if isinstance(var, Literal):
        return IgnoredGradient

    # ref id is used to distinguish the first
    # time seeing a variable; when we will
    # create an unversioned gradient symbol.
    if symbols.is_last_ref(var, ref.ref_id):
        return symbols.add_vjp(var)
    else:
        return symbols.add_vjp(var, ref.ref_id)

def connect_output_vjp(ref, symbols):

    # make lists for lists
    if isinstance(ref, ListRef):
        for r in ref:
            connect_output_vjp(r, symbols)
        return

    var = ref.symbol

    # bypass literal arguments
    if isinstance(var, Literal):
        return

    # accummulate the partials
    if not symbols.is_last_ref(var, ref.ref_id):
        var_f = symbols.get_vjp(var)
        var_p = symbols.get_vjp(var, ref.ref_id)
        # create a new symbol for the result, with the same name
        # because we intent to overwrite it.
        var_f2 = symbols.add_vjp(var)

        add.apl.create_node(dict(x1=var_f, x2=var_p), dict(y=var_f2))

def create_output_jvp(var, symbols):
    if isinstance(var, List):
        return [create_output_jvp(v, symbols) for v in var]

    if isinstance(var, Literal):
        raise RuntimError("This shall not happen, vjp is from an output which can never be a literal")

    return symbols.add_jvp(var)

def create_input_jvp(var, symbols):
    if isinstance(var, List):
        return [create_input_jvp(v, symbols) for v in var]

    if isinstance(var, Literal):
        return symbols.zero

    return symbols[var._jvp_name]

def create_input_vjp(var, symbols):
    if isinstance(var, List):
        return [create_input_vjp(v, symbols) for v in var]

    if isinstance(var, Literal):
        raise RuntimError("This shall not happen, vjp is from an output which can never be a literal")

    if not var._vjp_name in symbols:
        return symbols.zero
    return symbols[var._vjp_name]

def vjpmodel(tape):
    """ generate a vector jacobian product model based on a tape """
    model = type(tape.model)()
    symbols = SymbolCollection(model, tape)

    for var in tape.model._vout:
        model.input(symbols.add_vjp(var))

    for i, record in enumerate(tape[::-1]):
        p = record.node

        vjp_of_p = p.find_primitive_type(func='vjp')

        kwargs = prepare_opr_kwargs(record, model)
        kwout = {}
        # initialize 'v'
        for argname, var in p.varout.items():
            kwargs['_' + argname] = create_input_vjp(var, symbols)

        # create output vjps
        for argname, ref in p.varin.items():
            var_p = create_output_vjp(ref, symbols)

            if var_p is not None:
                kwout['_' + argname] = var_p

        # result unused; symbols directly connected via kwout
        vjp_of_p.create_node(kwargs=kwargs, kwout=kwout)

        # combine partial derivatives.
        for argname, ref in p.varin.items():
            connect_output_vjp(ref, symbols)

    # mark outputs
    for var in tape.model._vin:
        if not var._vjp_name in symbols:
            varout = symbols.zero
        else:
            varout = symbols[var._vjp_name]
        model.output(**{var._vjp_name : varout})

    return model

def jvpmodel(tape):
    """ generate a jacobian vector product model based on a tape """
    model = type(tape.model)()
    symbols = SymbolCollection(model, tape)

    for var in tape.model._vin:
        model.input(symbols.add_jvp(var))

    for i, record in enumerate(tape):
        p = record.node

        jvp_of_p = p.find_primitive_type(func='jvp')

        kwargs = prepare_opr_kwargs(record, model)
        kwout = {}
        # initialize 'v'
        for argname, ref in p.varin.items():
            jvp_var = create_input_jvp(ref.symbol, symbols)
            kwargs[argname + '_'] = jvp_var

        # create output symbols
        for argname, var in p.varout.items():
            jvp_var = create_output_jvp(var, symbols)
            kwout[argname + '_'] = jvp_var

        jvp_of_p.create_node(kwargs=kwargs, kwout=kwout)

    # mark outputs
    for var in tape.model._vout:
        if not var._jvp_name in symbols:
            varout = symbols.zero
        else:
            varout = symbols[var._jvp_name]
        model.output(**{var._jvp_name : varout})

    symbols.check(model)
    return model



from fnmatch import fnmatch
import re

class Entry(object):
    def __init__(self, operator, patterns):
        self.operator = operator
        self.patterns = self.make_patterns(patterns)
        self.str_patterns = patterns

    def make_patterns(self, patterns):
        d = []
        for arg, p in patterns.items():
            if isinstance(p, type):
                # special case if pattern is already an exact type
                # treat it like a string.
                p = p.__name__

            d.append((arg, re.compile(translate(p))))
        return dict(d)

    def match(self, operands):
        """ See if the entry matches the set of operands.

            Extra arguments in the operands are ignored.
        """

        # if all args in the pattern are supplied?
        if any(arg not in operands for arg in self.patterns):
            return None

        score_major = 0 # number of glob groups
        score_minor = 0 # length of glob pattern matches

        typenames = _gettypenames(operands)

        for arg in self.patterns:
            pattern = self.patterns[arg]
            typename = typenames[arg]

            r = pattern.match(typename)
            if not r:
                return None

            score_major += len(r.groups())
            score_minor += sum([len(x) for x in r.groups()])

        # prefer more specific rules
        return len(operands) - len(self.patterns), score_major, score_minor

    def __repr__(self):
        return "Entry: %s : %s" % (self.operator, self.str_patterns)

class Registry(object):
    """ A registory of operators and matching rules """
    def __init__(self):
        self._all = {}

    def match_all(self, name, operands):
        """ Find by name the operator entry that declares it can work on these operands """
        desc = self._all.get(name, [])
        r = []
        for i, entry in enumerate(desc[::-1]):
            score = entry.match(operands)
            if score is not None:
                r.append((score, i, entry, _gettypenames(operands)))

        r = sorted(r)
        return r

    def match(self, operator, operands, name=None):
        """ Find optimal operator entry that declares it can work on these operands """
        if name is None:
            name = _getname(operator)

        r = self.match_all(name, operands)

        if len(r) == 0:
            raise KeyError('no operator is found for %s' % str(operator))

        score, i, entry, types = r[0]
        return entry.operator

    def register(self, operator, patterns, name=None):
        """ Regsitor an operator entry that declares it can work on these operands matched by the pattern. 

            patterns: a dict of arg:fnmatch patterns -- must match from beginning to end on the name of the type;
        """
        if name is None:
            name = _getname(operator)

        desc = self._all.get(name, [])

        if patterns not in desc:
            desc.append(Entry(operator, patterns))

        self._all[name] = desc

def _gettypenames(operands):
    """ helper function to obtain type name dict from an input dict of operands """
    return dict([(arg, type(operand).__name__) for arg, operand in operands.items()])

def _getname(operator):
    """ helper function to infer operator name """
    if isinstance(operator, str):
        return operator
    return operator.__name__

def translate(pat):
    """Translate a shell PATTERN to a regular expression.
    There is no way to quote meta-characters.

    This differs from the one from fnmatch in that it converts each
    glob as a group. Such that we can count the number of chars that matched.

    """

    i, n = 0, len(pat)
    res = ''
    while i < n:
        c = pat[i]
        i = i+1
        if c == '*':
            res = res + r'(.*)'
        elif c == '?':
            res = res + r'(.)'
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j+1
            if j < n and pat[j] == ']':
                j = j+1
            while j < n and pat[j] != ']':
                j = j+1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pat[i:j].replace('\\','\\\\')
                i = j+1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res = r'%s([%s])' % (res, stuff)
        else:
            res = res + re.escape(c)
    return res + '\Z(?ms)'



from __future__ import print_function
from pprint import pprint
from vmad.lib import fastpm, linalg
import numpy
from vmad import autooperator

from vmad.testing import BaseScalarTest
from mpi4py import MPI
from numpy.testing import assert_allclose
from pmesh.pm import RealField, ComplexField

def create_bases(x):
    bases = numpy.eye(x.size).reshape([-1] + list(x.shape))
    if isinstance(x, RealField):
        pm = x.pm
        # FIXME: remove this after pmesh 0.1.36
        def create_field(pm, data):
            real = pm.create(type='real')
            real[...] = data
            return real
        return [create_field(pm, i) for i in bases]
    else:
        return [i for i in bases]

class Test_r2c_c2r(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')
    y = x.cnorm()
    x_ = create_bases(x)

    def model(self, x):
        c = fastpm.r2c(x)
        r = fastpm.c2r(c)
        return r

#    def teardown(self):
#        print(self.y_)

def transfer(k):
    k2 = sum(ki ** 2 for ki in k)
    mask = k2 == 0
    k2[mask] = 1
    r = 1 / k2
    r[mask] = 0
    return r

class Test_r2c_transfer_c2r(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')
    y = x.r2c().apply(lambda k, v: transfer(k) * v).c2r().cnorm()
    x_ = create_bases(x)

    def model(self, x):
        c = fastpm.r2c(x)
        c = fastpm.apply_transfer(c, tf=transfer)
        r = fastpm.c2r(c)
        return r

#    def teardown(self):
#        print(self.y_)

class Test_apply_digitized_amp_x(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4], BoxSize=4.0, comm=MPI.COMM_SELF)
    kedges = numpy.linspace(0, 2 * numpy.pi / 4.0 * 3, 8)
    tf = numpy.arange(len(kedges) - 1) + 1
    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')
    x_ = create_bases(x)
    y = NotImplemented

    def model(self, x):
        c = fastpm.r2c(x)
        digitizer = fastpm.apply_digitized.isotropic_wavenumber(self.kedges)
        c = fastpm.apply_digitized(c, tf=self.tf, digitizer=digitizer, kind='wavenumber')
        r = fastpm.c2r(c)
        return r

class Test_apply_digitized_phase_x(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[8, 8], BoxSize=4.0, comm=MPI.COMM_SELF)
    kedges = numpy.linspace(0, 2 * numpy.pi / 4.0 * 3, 8)
    tf = numpy.linspace(-1 * numpy.pi, 4 * numpy.pi, (len(kedges) - 1))
    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')
    x_ = create_bases(x)
    y = NotImplemented

    def model(self, x):
        c = fastpm.r2c(x)
        digitizer = fastpm.apply_digitized.isotropic_wavenumber(self.kedges)
        c = fastpm.apply_digitized(c, tf=self.tf, digitizer=digitizer, kind='wavenumber', mode='phase')
        r = fastpm.c2r(c)
        return r

class Test_apply_digitized_amp_tf(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=4.0, comm=MPI.COMM_SELF)
    # avoid nyquist!
    kedges = numpy.linspace(0, 2 * numpy.pi / 4.0 * 3, 8)
    y = NotImplemented
    x = numpy.arange(len(kedges) - 1)
    x_ = create_bases(x)
    atol = 1e-9
    def model(self, x):
        c0 = self.pm.generate_whitenoise(seed=300, unitary=True, type='complex', mean=1.0)
        digitizer = fastpm.apply_digitized.isotropic_wavenumber(self.kedges)
        c = fastpm.apply_digitized(c0, tf=x, digitizer=digitizer, kind='wavenumber')
        r = fastpm.c2r(c)
        return r

class Test_apply_digitized_phase_tf(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=4.0, comm=MPI.COMM_SELF)
    kedges = numpy.linspace(0, 2 * numpy.pi / 4.0 * 3, 8)
    y = NotImplemented
    x = numpy.linspace(-1 * numpy.pi, 4 * numpy.pi, (len(kedges) - 1))
    x_ = create_bases(x)
    atol = 1e-9
    def model(self, x):
        c0 = self.pm.generate_whitenoise(seed=300, unitary=True, type='complex', mean=1.0)
        digitizer = fastpm.apply_digitized.isotropic_wavenumber(self.kedges)
        c = fastpm.apply_digitized(c0, tf=x, digitizer=digitizer, kind='wavenumber', mode='phase')
        r = fastpm.c2r(c)
        return r

class Test_cdot(BaseScalarTest):
    to_scalar = staticmethod(linalg.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')
    x1 = pm.generate_whitenoise(seed=301, unitary=True, type='real')
    y = (x.cdot(x1)) ** 2 * pm.Nmesh.prod() ** -2.
    x_ = create_bases(x)

    def model(self, x):
        x1 = fastpm.r2c(self.x1)
        x = fastpm.r2c(x)
        return fastpm.cdot(x, x1)


class Test_paint_x(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    mesh = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    x = pm.generate_uniform_particle_grid(shift=0.1)
    y = (1 + mesh).cnorm()
    x_ = create_bases(x)

    epsilon = 1e-3

    def model(self, x):
        y = fastpm.paint(x, layout=None, mass=1.0, pm=self.pm)
        y = linalg.add(y, self.mesh) # biasing a bit to get non-zero derivatives.
        return y

class Test_decompose_paint_x(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    mesh = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    x = pm.generate_uniform_particle_grid(shift=0.1)
    y = (1 + mesh).cnorm()
    x_ = create_bases(x)

    epsilon = 1e-3

    def model(self, x):
        layout = fastpm.decompose(x, pm=self.pm)
        y = fastpm.paint(x, layout=layout, mass=1.0, pm=self.pm)
        y = linalg.add(y, self.mesh) # biasing a bit to get non-zero derivatives.
        return y

class Test_decompose_exchange(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    mesh = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    x = pm.generate_uniform_particle_grid(shift=0.1)
    y = (1 + mesh).cnorm()
    x_ = create_bases(x)

    epsilon = 1e-3

    def model(self, x):
        layout = fastpm.decompose(x, pm=self.pm)
        x1 = fastpm.exchange(x, layout)
        y1 = fastpm.paint(x1, mass=1.0, layout=None, pm=self.pm)
        y = linalg.add(y1, self.mesh) # biasing a bit to get non-zero derivatives.
        return y

class Test_paint_mass(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    mesh = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    pos = pm.generate_uniform_particle_grid(shift=0.1)
    x = numpy.ones(len(pos))
    y = x.sum()
    x_ = create_bases(x)

    epsilon = 1e-3

    def model(self, x):
        y = fastpm.paint(self.pos, layout=None, mass=x, pm=self.pm)
        return y

class Test_readout_x(BaseScalarTest):
    to_scalar = staticmethod(linalg.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    mesh = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    x = pm.generate_uniform_particle_grid(shift=0.5)
    y = (mesh.readout(x) ** 2).sum()
    x_ = create_bases(x)

    epsilon = 1e-3

    def model(self, x):
        y = fastpm.readout(self.mesh, x, layout=None)
        return y

class Test_readout_mesh(BaseScalarTest):
    to_scalar = staticmethod(linalg.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    pos = pm.generate_uniform_particle_grid(shift=0.5)
    y = (x.readout(pos) ** 2).sum()
    x_ = create_bases(x)

    epsilon = 1e-3

    def model(self, x):
        y = fastpm.readout(x, self.pos, layout=None)
        return y

class Test_lpt1(BaseScalarTest):
    to_scalar = staticmethod(linalg.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    pos = pm.generate_uniform_particle_grid(shift=0.5)
    y = NotImplemented
    x_ = create_bases(x)

    epsilon = 1e-4

    def model(self, x):
        dx1 = fastpm.lpt1(fastpm.r2c(x), q=self.pos, pm=self.pm)
        return dx1

class Test_lpt2src(BaseScalarTest):
    to_scalar = staticmethod(fastpm.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    y = NotImplemented
    x_ = create_bases(x)

    epsilon = 1e-4
    rtol = 1e-6
    def model(self, x):
        return fastpm.lpt2src(fastpm.r2c(x), pm=self.pm)

class Test_lpt(BaseScalarTest):
    to_scalar = staticmethod(linalg.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    pos = pm.generate_uniform_particle_grid(shift=0.5)
    y = NotImplemented
    x_ = create_bases(x)

    epsilon = 1e-4

    def model(self, x):
        dx1, dx2 = fastpm.lpt(fastpm.r2c(x), q=self.pos, pm=self.pm)
        return linalg.add(dx1, dx2)

class Test_nbody(BaseScalarTest):
    to_scalar = staticmethod(linalg.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    pos = pm.generate_uniform_particle_grid(shift=0.5)
    y = NotImplemented
    x_ = create_bases(x)

    epsilon = 1e-4
    def model(self, x):
        from nbodykit.cosmology import Planck15
        dx, p, f = fastpm.nbody(fastpm.r2c(x), q=self.pos, stages=[0.1, 0.5, 1.0], pm=self.pm, cosmology=Planck15)
        return linalg.stack([dx, p, f], axis=-1)

class Test_fastpm(BaseScalarTest):
    to_scalar = staticmethod(linalg.to_scalar)

    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')

    pos = pm.generate_uniform_particle_grid(shift=0.5)
    y = NotImplemented
    x_ = create_bases(x)

    epsilon = 1e-4
    def model(self, x):
        from nbodykit.cosmology import Planck15
        sim = fastpm.FastPMSimulation(pm=self.pm, cosmology=Planck15, stages=[0.1, 0.5, 1.0], q=self.pos)

        dx, p, f = sim.run(fastpm.r2c(x))
        return linalg.stack([dx, p, f], axis=-1)


def test_add_scalar():
    from vmad import autooperator
    from vmad.core.stdlib import eval
    from numpy.testing import assert_array_equal
    pm = fastpm.ParticleMesh(Nmesh=[4, 4, 4], BoxSize=8.0, comm=MPI.COMM_SELF)

    x = pm.generate_whitenoise(seed=300, unitary=True, type='real')
    x[...] = 1.0

    @autooperator('x, n->y')
    def func(x, n):
        a = linalg.take(n, 0, axis=0)
        a = linalg.broadcast_to(a, eval(x, lambda x : x.shape))
        return x + a

    n = numpy.array([3, 4])
    (y, ), (_x, _n) = (func.build().compute_with_vjp(init=dict(x=x, n=n), v=dict(_y=x)))
    assert_array_equal(y, 4)
    assert_array_equal(_x, 1)
    assert_array_equal(_n, (64, 0))



import inspect
from collections import OrderedDict

from .error import UnpackError, OverwritePrecaution, MissingArgument, BrokenPrimitive, BadArgument
from .node import Node

from .symbol import Symbol, assymbol, BaseSymbol, Literal, List
from .model import Model

def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

class Primitive:
    """ Primitives are building blocks of models.

        Calling a primitive creates a node on a model.

        A primitive object for an operator is generated
        and attached to the operators object via the `operator` decorator.

    """

    def __init__(self, func, opr, ain, aout, argnames, outnames, impl, record_impl):
        self.name = opr.prototype.__name__ + '-' + func
        self.func = func
        self.operator = opr
        self.ain = ain
        self.aout = aout
        self.argnames = argnames
        self.outnames = outnames

        self.impl = impl
        self._default_kwargs = get_default_args(self.impl)
        self.record_impl = record_impl

        # a few others are created in make_primitive

    def __eq__(self, other):
        """ If two primitives are the same, they must be for the same operator
            and the same function type. """
        if self.operator is not other.operator: return False
        if self.func != other.func: return False
        return True

    # When there are multiple return values, or if the return variable(s)
    # are fed on the calling arguments, it behaves like a list,
    # and must be unpacked. A strange error (during resolving) is raised if not
    # unpacked.

    def create_node(self, kwargs, kwout=None, stacklevel=-2):
        """ Creating a node and append it to the model's execution graph

            The model is inferred from the input arguments

            stacklevel is the adjustment for stacklevel
            -1 is the caller. -2 is the caller of the caller

            kwargs are the input arguments; values are symbols.
            kwout are the output arguments, if an output argument is not
            given, a new symbol will be created.
        """
        if kwout is None: # generate output arguments.
            kwout = {}
        # remember the frame info
        previous_frame = inspect.currentframe()

        while stacklevel < 0:
            previous_frame = previous_frame.f_back
            stacklevel = stacklevel + 1

        _frameinfo = inspect.getframeinfo(previous_frame)

        node = Node(self, _frameinfo)

        # FIXME: this is tricky.
        # we guarentee prepending all necessary sub-models
        # necessary for this node before this
        # node; therefore this shall always work as long
        # as we always build a model this way.
        # the algorithm here does not explicitly say this.

        # gather models
        models = self._walk_models(kwargs, kwout)
        # consolidate
        model = Model.consolidate(models)
        basename = model.unique_name(self.name)

        for argname in self.argnames:
            if argname in self.ain:
                var = assymbol(kwargs[argname])
            else:
                # if it is not provided (e.g. default value defined in the prototype is used)
                if argname not in kwargs:
                    continue
                # not a autodiff argument, must be a Literal
                # but we still allow passing in a symbol
                # just ignore its gradient propagation.
                var = assymbol(kwargs[argname])

            # checking symbol references
            #print(basename, var.name, id(var), id(model.get(var.name)))

            ref = var._add_reference(node)
            node._varin[argname] = ref

        for argname in self.outnames:
            if not argname in kwout:
                # if a name is not supplied, generate a name
                varname = basename + '-' + argname
                var = Symbol(varname, model=model)
            else:
                var = assymbol(kwout[argname])

                # already given a symbol, overwrite it
                # but this doesn't work for gradients / tape
                # so we die here
                _check_var_references(var)

            node._varout[argname] = var

        # append node to the model.
        model.append(node)

        # return the output symbols
        vout = [node.varout[argname] for argname in self.outnames]

        return OrderedDict([(argname, node.varout[argname]) for argname in self.outnames])

    def _parse_args(self, args, kwargs):
        """ map arguments give as args and kwargs to argnames. Used in __call__

        """

        kwout = {}
        # first attempt to map args into kwargs
        if len(args) > len(self.argnames):
            raise BadArgument("Argument list longer than total number of args")

        for argname, arg in zip(self.argnames, args):
            if argname in kwargs:
                raise BadArgument("argument %s already provided as keyword" % argname)
            kwargs[argname] = arg

        for argname, arg in self._default_kwargs.items():
            if argname not in kwargs:
                kwargs[argname] = arg

        for argname in list(kwargs.keys()):
            if argname in self.outnames:
                if argname not in self.argnames:
                    import warnings
                    warnings.warn("Supplying keyword argument to an operator for an output. Prefer to use the verbose .create_node API instead of this", DeprecationWarning, stacklevel=3)
                    kwout[argname] = kwargs.pop(argname)

            elif argname not in self.argnames:
                raise BadArgument("Argument %s is passed in , but not declared as an argument" % argname)

        for argname in self.argnames:
            if argname not in kwargs:
                raise MissingArgument("Argument %s is missing." % argname)

        return kwargs, kwout


    def _walk_models(self, kwargs, kwout):
        models = list()

        q = []
        for argname in self.argnames:
            var = kwargs[argname]
            q.append(var)

        for argname in self.outnames:
            # will automatically generate kwout, and they will be properly anchored
            if argname not in kwout: continue
            var = kwout[argname]
            q.append(var)

        for var in q:
            m1 = _infer_models(var)
            _join_models(models, m1)
        return models



def _check_var_references(var):
    from .symbol import List
    if isinstance(var, List):
        for v in var:
            _check_var_references(v)
        return

    if var._has_reference():
        raise OverwritePrecaution("Overwritting used symbols is not supported. Because it breaks vjp.")

def _join_models(models, m1):
    for m in m1:
        if m not in models:
            models.append(m)

def _infer_models(var):

    if isinstance(var, Symbol):
        model = var._model
        if model is not None:
            return [model]
        else:
            return []

    if isinstance(var, (List, list, tuple)):
        models = []
        for v in var:
            m1 = _infer_models(v)
            _join_models(models, m1)

        return models

    return []

